# Copyright 2016 Virantha Ekanayake All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import os, shutil, logging

from target import TargetBase

class FileCopy(TargetBase):

    def __init__(self):
        pass

    def check_duplicates(self, images):
        """ This is easy, since all the functionality is built into the source image file
            object
        """
        print("Checking for duplicates")
        images_1, images_2 = itertools.tee(images)
        for total,img in enumerate(images_1):
            if img.is_duplicate():
                img.dup = True

        n_dups = [i for i in images_2 if i.dup]
        print('Found {} duplicates out of {} images'.format(len(n_dups), total+1))


    def _get_unique_filename_suffix(self, filename):
        dirname = os.path.dirname(filename)
        fn_with_ext = os.path.basename(filename)
        fn, ext = os.path.splitext(fn_with_ext)
        
        suffix = 1
        if not os.path.exists(filename):  # Unique, no target filename conflict
            return filename
        else:
            while os.path.exists(os.path.join(dirname, fn+'_'+str(suffix)+ext)):
                suffix += 1
            return (os.path.join(dirname, fn+'_'+str(suffix)+ext))

    def execute_copy(self, images):

        skip_count = 0
        print("Copying and sorting files")
        for total, img in enumerate(images):
            if img.dup: 
                skip_count+=1
                continue
            srcfn = img.srcpath
            tgtfn = img.tgtpath
            tgtdir = os.path.dirname(tgtfn)
            tgtfn = self._get_unique_filename_suffix(tgtfn)
            logging.info("Copying %s to %s" % (srcfn, tgtfn))
            if not os.path.exists(tgtdir):
                logging.info("Creating directory {}".format(tgtdir))
                os.makedirs(tgtdir)

            shutil.copyfile(srcfn, tgtfn)
        print ("Skipped {} duplicate files".format(skip_count))
        print ("Copied {} files".format(total+1-skip_count))




# -*- coding: utf-8 -*-
# Copyright 2016 Virantha Ekanayake All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys, os
import logging

import yaml, pprint
import flickrapi
import urllib.request
from xml.etree import ElementTree
from tqdm import tqdm
import itertools, dateparser, time

from target import TargetBase




class FileWithCallback(object):
    def __init__(self, filename):
        self.file = open(filename, 'rb')
        # the following attributes and methods are required
        self.len = os.path.getsize(filename)
        self.fileno = self.file.fileno
        self.tell = self.file.tell
        self.tqdm = tqdm(total=self.len, ncols=60,unit_scale=True, unit='B')

    def read(self, size):
        self.tqdm.update(size)
        return self.file.read(size)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.tqdm.close()


class FlickrMedia(object):
    def __init__(self, json_dict):
        self.json_dict = json_dict
        self.title = json_dict['title']
        self.photoid = json_dict['id']

        dt = json_dict['datetaken']
        self.datetime_taken = dateparser.parse(dt, date_formats=['%Y-%m-%d %H:%M:%S']) 


class PhotoSet(object):

    def __init__(self, json_dict):
        self.json_dict = json_dict
        self.title = json_dict['title']['_content']
        self.setid = json_dict['id']
        self.photos = None

class Photo(object):
    def __init__(self, photo_element):
        """Construct a photo object out of the XML response from Flickr"""
        attrs = { 'farm': 'farmid', 'server':'serverid','id':'photoid','secret':'secret'}
        for flickr_attr, py_attr in attrs.items():
            setattr(self, py_attr, photo_element.get(flickr_attr))
        
    def _construct_flickr_url(self):
        url = "http://farm%s.staticflickr.com/%s/%s_%s_b.jpg" % (self.farmid,self.serverid, self.photoid, self.secret)
        return url

    def download_photo(self, dirname, cache=False, tgt_filename=None):
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        tgt = os.path.join(dirname, "%s.jpg" % self.photoid)
        if cache:
            if os.path.isfile(tgt):
                return tgt
        urllib.request.urlretrieve(self._construct_flickr_url(), tgt)
        return tgt
        
class Flickr(TargetBase):

    def __init__(self):
        self.set_keys(*self.read_keys())
        self.get_auth2()
        # Might as well get all the photosets at this point as we'll need them
        self.photosets = self._get_photosets()


    def read_keys(self):
        """
            Read the flickr API key and secret from a local file
        """
        with open("flickr_api.yaml") as f:
            api = yaml.load(f)
        return (api["key"], api["secret"])

    def set_keys(self, key, secret):
        self.api_key = key
        self.api_secret = secret

    def get_auth2(self):
        print("Authenticating to Flickr")
        self.flickr = flickrapi.FlickrAPI(self.api_key, self.api_secret)
        self.flickr.authenticate_via_browser(perms='write')
        print("Authentication succeeded")
        return 

    def get_tagged(self, tags, count, download_dir="photos"):
        """ Get photos with the given list of tags
        """
        print ("connecting to flickr, and getting %d photos with tags %s" % (count, tags))
        x = self.flickr.photos_search(api_key = self.api_key, user_id="me", tags=','.join(tags), per_page=count)
        photos = self._extract_photos_from_xml(x)
        photo_filenames = self._sync_photos(photos, download_dir)
        print("Found %d photos" % len(photos))
        return photo_filenames


    def _sync_photos(self, photos, download_dir="photos", clean_up=False):
        """
            Connect to flickr, and for each photo in the list, download.
            Then, if delete photos that are present locally that weren't present in the list of photos.

            :returns: List of filenames downloaded
        """
        photo_filenames = []
        photo_count = len(photos)
        for i,photo in enumerate(photos):
            print("[%d/%d] Downloading %s from flickr" % (i,photo_count,photo.photoid))
            filename = photo.download_photo(download_dir, cache=True)
            photo_filenames.append(filename)

        # Now, go through and clean up directory if required
        
        if clean_up:
            photo_file_list = ["%s.jpg" % (x.photoid) for x in photos]
            for fn in os.listdir(download_dir):
                full_fn = os.path.join(download_dir, fn)
                if os.path.isfile(full_fn):
                    if not fn in photo_file_list:
                        print ("Flickr sync: Deleting file %s" % fn)
                        os.remove(full_fn)

        return photo_filenames

    def _extract_photos_from_xml(self, xml):
        photos = []
        for i in xml.iter():
            if i.tag == 'rsp':
                # the response header.  stat member should be 'ok'
                if i.get('stat') == 'ok':
                    continue
                else:
                    # error, so just break
                    break
            if i.tag == 'photo':
                photos.append(Photo(i))
        return photos

    def get_recent(self,count, download_dir="photos"):
        """ get the most recent photos
        """
        print ("connecting to flickr, and getting most recent %d photos" % count)
        x = self.flickr.people_getphotos(api_key = self.api_key, user_id="me",per_page=count)
        #x = self.flickr.photos_search(api_key=self.api_key,"me")

        photos = self._extract_photos_from_xml(x)
        photo_filenames = self._sync_photos(photos, download_dir)
        return photo_filenames


    def _get_photosets(self):
        print("Getting photosets from Flickr")
        resp = self.flickr.photosets.getList(format='parsed-json')
        photosets = {}
        for photoset in resp['photosets']['photoset']:
            p = PhotoSet(photoset)
            photosets[p.title] = p #TODO: Possible issue here because multiple photosets could have same title.  Oh well
        return photosets


    def _get_photos_in_album(self, album_name, cached=False):
        photoset = self.photosets[album_name]
        albumid = photoset.setid
        if not photoset.photos or not cached:
            resp = self.flickr.photosets.getPhotos(photoset_id=albumid, extras='date_taken', format='parsed-json')
            photos = {}
            for p in resp['photoset']['photo']:
                myphoto = FlickrMedia(p)
                photos[myphoto.title] = myphoto
            photoset.photos = photos
        return photoset.photos


    def _upload_file(self, filename):
        with FileWithCallback(filename) as f:
            resp = self.flickr.upload(filename=filename, fileobj=f, is_public=0)
            photoid = resp.find('photoid').text
            return photoid


    def _create_new_album(self, album_name, first_photo_filename):

        # First, we need to upload a dummy photo
        photoid = self._upload_file(first_photo_filename)

        resp = self.flickr.photosets.create(title=album_name, primary_photo_id=photoid, format='parsed-json')
         
        albumid = resp['photoset']['id']
        resp = self.flickr.photosets.getInfo(photoset_id=albumid, format='parsed-json')

        return (photoid, resp['photoset'])

        
    def _add_photo_to_album(self, photoid, albumid):
        #tqdm.write("Adding {} to {} ".format(photoid, albumid))
        self.flickr.photosets.addPhoto(photoset_id=albumid, photo_id=photoid)

    def _is_duplicate(self, image):
        album_name = image.tgtdatedir

        if not album_name in self.photosets:
            return False
        else:
            photos = self._get_photos_in_album(album_name, cached=True)
            image_title = os.path.basename(image.filename)
            if not image_title in photos:  # If photo with same title is not found, then no duplicates
                return False
            else:
                # Same title, but let's check the date too, to be sure
                #tqdm.write('{} has local date {}, and flickr date {}'.format(image_title, image.datetime_taken, photos[image_title].datetime_taken))
                if photos[image_title].datetime_taken != image.datetime_taken:
                    return False
                else:
                    return True


    def check_duplicates(self, images):
        print("Checking for duplicates in Flickr")
        images_1, images_2 = itertools.tee(images)
        for total,img in enumerate(images_1):
            if self._is_duplicate(img):
                img.flickr_dup = True

        n_dups = [i for i in images_2 if i.flickr_dup]
        print('Found {} duplicates out of {} images'.format(len(n_dups), total+1))


    def execute_copy(self, images):

        for img in images:
            if img.flickr_dup: continue
            album_name = img.tgtdatedir
            if album_name not in self.photosets:
                # Need to create album
                tqdm.write('Creating new album %s' % album_name)
                photoid, album_dict = self._create_new_album(album_name, img.srcpath)
                p = PhotoSet(album_dict)
                self.photosets[p.title] = p
            else:
                photoid = self._upload_file(img.srcpath)
                self._add_photo_to_album(photoid, self.photosets[album_name].setid)

            tqdm.write("Adding {} to {} ".format(img.filename, album_name))
            # Now, make sure we set the date-taken manually if no exif information
            if img.exif_timestamp_missing:
                dt = img.datetime_taken.strftime('%Y-%m-%d %H:%M:%S')
                tqdm.write('Manually setting date on video {} to {}'.format(img.filename, dt))
                self.flickr.photos.setDates(photo_id=photoid, date_taken=dt)



def main():
    #logging.basicConfig(level=logging.DEBUG, format='%(message)s')
    script = Flickr()
    #script.get_recent(10)
    #script.upload('test.jpg')
    script.flickr.photos.setDates(photoid='30735623495', date_taken='2016-06-24 10:12:02')


if __name__ == '__main__':
    main()





# Copyright 2016 Virantha Ekanayake All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

Usage:
    photokeeper.py [options] SOURCE_DIR examine
    photokeeper.py [options] SOURCE_DIR TARGET_DIR [dedupe] file
    photokeeper.py [options] SOURCE_DIR [dedupe] flickr
    photokeeper.py [options] SOURCE_DIR TARGET_DIR [dedupe] file flickr
    photokeeper.py [options] SOURCE_DIR TARGET_DIR all
    photokeeper.py --conf=FILE
    photokeeper.py -h

Arguments:
    SOURCE_DIR  Source directory of photos
    TARGET_DIR  Where to copy the image files
    all         Run all steps in the flow (%s)
%s

Options:
    -h --help        show this message
    -v --verbose     show more information
    -d --debug       show even more information
    --conf=FILE      load options from file

"""

from docopt import docopt
import yaml
import sys, os, logging, shutil, datetime, pprint, filecmp
from collections import OrderedDict, defaultdict
from schema import Schema, And, Optional, Or, Use, SchemaError
import piexif, dateparser

from tqdm import tqdm
from flickr import Flickr
from filecopy import FileCopy

from version import __version__
from utils import ordered_load, merge_args

"""
   
.. automodule:: photokeeper
    :private-members:
"""
class ImageFile(object):
    def __init__(self, srcdir, filename, tgtbasedir, tgtdatedir, datetime_taken, exif_timestamp_missing=False):
        self.srcdir = srcdir
        self.filename = filename
        self.tgtbasedir = tgtbasedir
        self.tgtdatedir = tgtdatedir
        self.datetime_taken = datetime_taken
        self.dup = False
        self.flickr_dup = False
        self.exif_timestamp_missing = exif_timestamp_missing
        #print("adding {} with datetime {}".format(filename, datetime_taken.strftime('%Y-%m-%d %H:%M:%S')))
        pass

    @property
    def srcpath(self):
        return os.path.join(self.srcdir, self.filename)
    
    @property
    def tgtpath(self):
        return os.path.join(self.tgtbasedir, self.tgtdatedir, self.filename)

    def is_duplicate(self, shallow_compare = True):
        # First, see if there is a file already there
        if not os.path.exists(self.tgtpath):
            return False
        elif os.path.getsize(self.srcpath) != os.path.getsize(self.tgtpath):
            return False
        #elif not filecmp.cmp(self.srcpath, self.tgtpath, shallow_compare):  # This is too slow over a network share
        #   return False
        return True


class PhotoKeeper(object):
    """
        The main clas.  Performs the following functions:

    """

    def __init__ (self):
        """ 
        """
        self.args = None
        self.flow = OrderedDict([ ('examine', 'Examine EXIF tags'),
                                  ('dedupe', 'Only select files not already present in target directory'),
                                  ('flickr', 'Upload to flickr'),
                                  ('file',    'Copy files'),
                      ])
        self.images = []




    def get_options(self, argv):
        """
            Parse the command-line options and set the following object properties:

            :param argv: usually just sys.argv[1:]
            :returns: Nothing

            :ivar debug: Enable logging debug statements
            :ivar verbose: Enable verbose logging
            :ivar config: Dict of the config file

        """
        padding = max([len(x) for x in self.flow.keys()]) # Find max length of flow step names for padding with white space
        docstring = __doc__ % (#'|'.join(self.flow), 
                              ','.join(self.flow.keys()),
                              '\n'.join(['    '+k+' '*(padding+4-len(k))+v for k,v  in self.flow.items()]))
        args = docopt(docstring, version=__version__)

        # Load in default conf values from file if specified
        if args['--conf']:
            with open(args['--conf']) as f:
                conf_args = yaml.load(f)
        else:
            conf_args = {}
        args = merge_args(conf_args, args)
        print (args)
        schema = Schema({
            'SOURCE_DIR': Or(os.path.isdir, error='Source directory does not exist'),
            'TARGET_DIR': Or(lambda x: x is None, os.path.isdir, error='Destination directory does not exist'),
            object: object
            })
        try:
            args = schema.validate(args)
        except SchemaError as e:
            exit(e)

        print (args)
        if args['all'] == 0:
            for f in list(self.flow):
                if args[f] == 0: del self.flow[f]
            logging.info("Doing flow steps: %s" % (','.join(self.flow.keys())))

        self.src_dir = args['SOURCE_DIR']
        self.tgt_dir = args['TARGET_DIR']
        if self.tgt_dir:
            assert os.path.abspath(self.src_dir) != os.path.abspath(self.tgt_dir), 'Target and source directories cannot be the same'

        if args['--debug']:
            logging.basicConfig(level=logging.DEBUG, format='%(message)s')
        elif args['--verbose']:
            logging.basicConfig(level=logging.INFO, format='%(message)s')   

        self.args = args # Just save this for posterity


    def get_file_count(self, _dir):
        cnt = 0
        for root, dirs, files in os.walk(_dir):
            cnt += len(files)
        return cnt


    def examine_files(self, img_dir):
        counts = defaultdict(int)
        dt_format = '%Y-%m-%d'
        pp = pprint.PrettyPrinter(indent=4)
        raw_count = self.get_file_count(img_dir)
        print("Examining {} files in {}".format(raw_count, img_dir))
        with tqdm(total=raw_count, ncols=80, unit='file') as progress:
            for root, dirs, files in os.walk(img_dir):
                for fn in files:
                    if fn.startswith('.'): continue
                    filename = os.path.join(root, fn)
                    progress.update(1)
                    try:
                        exif_timestamp_missing = False
                        tags_dict = piexif.load(filename)
                        image_date = tags_dict['0th'][piexif.ImageIFD.DateTime]
                        # Why am I even using dateparser if it can't parse this??
                        image_datetime = dateparser.parse(image_date.decode('utf8'), date_formats=['%Y:%m:%d %H:%M:%S']) 
                    except (KeyError, ValueError) as e:

                        logging.info('IGNORED: %s is not a JPG or TIFF' % (filename))
                        file_mod_time = os.path.getmtime(filename)
                        image_datetime = datetime.datetime.fromtimestamp(file_mod_time)
                        logging.info('Using %s ' % (image_datetime))
                        exif_timestamp_missing = True  # Need to mark this since we don't have EXIF and Flickr doesn't honor file date for date-taken
                    
                    image_datetime_text = image_datetime.strftime(dt_format)
                    counts[image_datetime_text] += 1
                    self.images.append(ImageFile(os.path.dirname(filename), os.path.basename(filename), self.tgt_dir, image_datetime_text, image_datetime, exif_timestamp_missing))

        counts = dict(counts)
        total = sum(counts.values())
        print('Found images from {} days'.format(len(counts)))
        pp.pprint (counts)
        
        print('Total images: {}'.format(total))


    def _get_unique_filename_suffix(self, filename):
        dirname = os.path.dirname(filename)
        fn_with_ext = os.path.basename(filename)
        fn, ext = os.path.splitext(fn_with_ext)
        
        suffix = 1
        if not os.path.exists(filename):  # Unique, no target filename conflict
            return filename
        else:
            while os.path.exists(os.path.join(dirname, fn+'_'+str(suffix)+ext)):
                suffix += 1
            return (os.path.join(dirname, fn+'_'+str(suffix)+ext))
            


    def all_images(self):
        n = len(self.images)
        with tqdm(total=n, ncols=80, unit='file') as progress:
            for i, img in enumerate(self.images):
                yield img
                progress.update(1)

    def go(self, argv):
        """ 
            The main entry point into PhotoKeeper

            #. Do something
            #. Do something else
        """
        # Read the command line options
        self.get_options(argv)
        self.examine_files(self.src_dir)
        for photo_target, TargetClass in [('file', FileCopy), ('flickr', Flickr)]:
            if photo_target in self.flow:
                f = TargetClass()
                if 'dedupe' in self.flow:
                    f.check_duplicates(self.all_images())
                f.execute_copy(self.all_images())

def main():
    script = PhotoKeeper()
    script.go(sys.argv[1:])

if __name__ == '__main__':
    main()




import photokeeper.photokeeper as P
import pytest
import os
import logging

import smtplib
from mock import Mock
from mock import patch, call
from mock import MagicMock
from mock import PropertyMock


class Testphotokeeper:

    def setup(self):
        self.p = P.PhotoKeeper()



import yaml
from collections import OrderedDict

def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
    """ Helper function to allow yaml load routine to use an OrderedDict instead of regular dict.
        This helps keeps things sane when ordering the runs and printing out routines
    """
    class OrderedLoader(Loader):
        pass
    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))
    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)
    return yaml.load(stream, OrderedLoader)


def merge_args(conf_args, orig_args):
    """ Return new dict with args, and then conf_args merged in.
        Make sure that any keys in conf_args are also present in args
    """
    args = {}
    for k in conf_args.keys():
        if k not in orig_args:
            print("ERROR: Configuration file has unknown option %s" % k)
            sys.exit(-1)

    args.update(orig_args)
    args.update(conf_args)
    return args



from __future__ import print_function
from setuptools import setup, find_packages

import io
import os
import photokeeper
from photokeeper.version import __version__
from setuptools import Command

class PyTest(Command):
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        import sys,subprocess
        cwd = os.getcwd()
        os.chdir('test')
        errno = subprocess.call([sys.executable, 'runtests.py'])
        os.chdir(cwd)
        raise SystemExit(errno)

def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)

packages = find_packages(exclude="tests")

long_description = read('README.rst')

with open("requirements.txt") as f:
    required = f.read().splitlines()

setup (
    name = "photokeeper",
    version = __version__,
    description="Keep your photos in sync with Flickr",
    license = "ASL 2.0",
    long_description = long_description,
    author="Virantha N. Ekanayake",
    author_email="virantha@gmail.com", # Removed.
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        ],
    package_data = {'': ['*.xml']},
    zip_safe = True,
    include_package_data = True,
    packages = packages,
    install_requires = required,
    entry_points = {
            'console_scripts': [
                    'photokeeper = photokeeper.photokeeper:main'
                ],
        },
    options = {
	    "pyinstaller": {"packages": packages}
	    },
    cmdclass = {'test':PyTest}

)



# Copyright 2016 Virantha Ekanayake All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

class TargetBase(abc.ABC):

    @abc.abstractmethod
    def check_duplicates(self, images):
        """ Go through images and mark a property in each image if it's a duplicate in the target repository
            Each plugin target type can mark a different 'dup' property
        """

    @abc.abstractmethod
    def execute_copy(self, images):
        """ Take the source image files and copy/upload them to the target repository
        """
